// map.cpp - 'A Quest Too Far'
//
// Copyright 2010 Geoffrey White
//
// see readme.txt for more details.

#include "map.h"
#include "game.h"
#include "creature.h"
#include "item.h"
#include "rng.h"
#include <assert.h>

//#define DEBUG_LEVELGEN

// --- directional stuff ---

static int dir_xs[8] = {0, 1, 1, 1, 0, -1, -1, -1};
static int dir_ys[8] = {-1, -1, 0, 1, 1, 1, 0, -1};

// --- level types ---

struct level_info
{
	char *introduction;
	char_style rock_style;
	int rooms_gen, room_min, room_max;
	int tunnels_gen, min_area;
	int monsters_min, monsters_max;
};
level_info level_infos[LT_NUM] =
//		 ########################################################
{	// LT_ENTRY
	{	"You enter the cave.  It's cold and dark, and the floor is slimy.  Brings back memories...",
		{224, 255, 224, 0, 0, 0},
		500, 3, 8,		// largeish rooms (rounded)
		300,
		700,
		10, 15			// fewer monsters
	},
	// LT_TUNNELS
	{	"There's a foul smell to these narrow, earthy tunnels.",
		{255, 224, 192, 0, 0, 0},
		50, 2, 4,		// very few, very small rooms
		750,			// lots of tunnels
		250,
		15, 20
	},
	// LT_CRYPT
	{	"There's a chill to the air in this dusty level.",
		{224, 224, 255, 0, 0, 0},
		3000, 2, 6,		// lots of smallish rooms
		300,			// lots of tunnels
		600,
		15, 20
	},
	// LT_GOBLINHALLS
	{	"The sound of distant chatter fills these halls.",
		{255, 255, 224, 0, 0, 0},
		2000, 4, 14,	// lots of large rooms
		200,
		900,
		30, 40			// double number of monsters
	},
	// LT_LAIR
	{	"There's a wicked cackling sound in the distance.",
		{192, 192, 192, 0, 0, 0},
		300, 2, 8,		// some unchecked rooms
		100,			// fewer tunnels
		700,
		15, 20
	},
};

// --- map ---

map :: map()
{
	int x, y;

	for (y = 0; y < MAP_HEIGHT; y++)
	{
		for (x = 0; x < MAP_WIDTH; x++)
		{
			tiles[x][y].display_ch = ' ';
			tiles[x][y].display_style = char_default_style;
			tiles[x][y].is_fx = false;
			tiles[x][y].vis = V_HIDDEN;
			tiles[x][y].my_creature = NULL;
		}
	}
}

map :: ~map()
{
	clear_creatures();
	clear_items();
}

void map :: clear_items()
{
	item *_item;
	int x, y;

	// delete all items on the map
	for (y = 0; y < MAP_HEIGHT; y++)
	{
		for (x = 0; x < MAP_WIDTH; x++)
		{
			while (!tiles[x][y].my_items.empty())
			{
				_item = tiles[x][y].my_items.front();
				delete _item;
				tiles[x][y].my_items.pop_front();
			}
		}
	}
}

void map :: clear_creatures()
{
	int x, y;

	// delete all creatures on the map
	for (y = 0; y < MAP_HEIGHT; y++)
	{
		for (x = 0; x < MAP_WIDTH; x++)
		{
			if (tiles[x][y].my_creature != NULL)
			{
				delete tiles[x][y].my_creature;
				tiles[x][y].my_creature = NULL;
			}
		}
	}
}

bool map :: add_room(level_type level)
{
	int x, y, w, h, x2, y2;
	bool dig, ring_room;

	// place the room
	x = random(1, MAP_WIDTH - 3);
	y = random(1, MAP_HEIGHT - 3);
	w = random(level_infos[level].room_min, level_infos[level].room_max);
	h = random(level_infos[level].room_min, level_infos[level].room_max);
	if ((x + w - 1 >= MAP_WIDTH - 1) ||
		(y + h - 1 >= MAP_HEIGHT - 1))
	{
		return false;
	}

	if ((level == LT_LAIR) && (random(15) == 0))
	{
		// occasional overlapping rooms
	} else {
		// check the room
		for (y2 = y - 1; y2 < y + h + 1; y2++)
		{
			for (x2 = x - 1; x2 < x + w + 1; x2++)
			{
				if (!tiles[x2][y2].is_solid)
				{
					return false;
				}
			}
		}
	}

	ring_room = false;
	if ((level == LT_GOBLINHALLS) && (random(3) == 0))
	{
		ring_room = true;
	}

	// add the room
	for (y2 = y; y2 < y + h; y2++)
	{
		for (x2 = x; x2 < x + w; x2++)
		{
			dig = true;

			if (ring_room)
			{
				if ((x2 >= x + 2) && (x2 < x + w - 2) &&
					(y2 >= y + 2) && (y2 < y + h - 2))
				{
					dig = false;
				}
			}
			if (level == LT_ENTRY)
			{
				// rounded rooms
				if (((x2 == x) || (x2 == x + w - 1)) &&
					((y2 == y) || (y2 == y + h - 1)) &&
					(random(3) != 0))
				{
					dig = false;
				}
			}

			if (dig)
			{
				tiles[x2][y2].display_ch = '.';
				tiles[x2][y2].is_solid = false;
			}
		}
	}

	return true;
}

struct point
{
	int x, y;
};

bool map :: add_tunnel(level_type level)
{
	std::vector<point> points;
	std::vector<point>::iterator it;
	point p;
	int dir;

	// start the tunnel
	p.x = random(1, MAP_WIDTH - 2);
	p.y = random(1, MAP_HEIGHT - 2);
	dir = random(4) * 2;
	if (level == LT_TUNNELS)
	{
		// allow diagonal tunnels
		dir = random(8);
	}

	// tunnels must start in a room
	if (tiles[p.x][p.y].is_solid)
	{
		return false;
	}

	// go to the edge of the room
	while (!tiles[p.x][p.y].is_solid)
	{
		p.x += dir_xs[dir];
		p.y += dir_ys[dir];
		if ((p.x <= 0) || (p.x >= MAP_WIDTH - 1) ||
			(p.y <= 0) || (p.y >= MAP_HEIGHT - 1))
		{
			// going off map
			return false;
		}
	}

	// trace the tunnel
	while (tiles[p.x][p.y].is_solid)
	{
		points.push_back(p);

		// chance of turning mid-tunnel
		if (level == LT_TUNNELS)
		{
			if (random(10) == 0)
			{
				dir = (dir + 1) % 8;
			}
			if (random(10) == 0)
			{
				dir = (dir + 7) % 8;
			}
		} else {
			if (random(40) == 0)
			{
				dir = (dir + 2) % 8;
			}
			if (random(40) == 0)
			{
				dir = (dir + 6) % 8;
			}
		}

		if ((!tiles[p.x + dir_xs[(dir + 2) % 8]][p.y + dir_ys[(dir + 2) % 8]].is_solid) || // 90 degrees right
			(!tiles[p.x + dir_xs[(dir + 6) % 8]][p.y + dir_ys[(dir + 6) % 8]].is_solid)) // 90 degrees left
		{
			// clipping another tunnel
			return false;
		}

		p.x += dir_xs[dir];
		p.y += dir_ys[dir];
		if ((p.x <= 0) || (p.x >= MAP_WIDTH - 1) ||
			(p.y <= 0) || (p.y >= MAP_HEIGHT - 1))
		{
			// going off map
			return false;
		}
	}

	// add the tunnel
	for (it = points.begin(); it != points.end(); it++)
	{
		p = *it;
		tiles[p.x][p.y].display_ch = '.';
		tiles[p.x][p.y].is_solid = false;
	}

	return true;
}

bool map :: add_player(game &the_game, creature *the_hero)
{
	int x, y, i;

	assert(the_hero != NULL);

	for (i = 0; i < 100; i++)
	{
		x = random(1, MAP_WIDTH - 2);
		y = random(1, MAP_HEIGHT - 2);
		if ((!tiles[x][y].is_solid) &&
			(tiles[x][y].my_creature == NULL))
		{
			the_hero->move_to(this, x, y);
			return true;
		}
	}

	return false;
}

bool map :: add_stairs(game &the_game, level_type level)
{
	int x, y, sx, sy, sspots;

	assert(the_game.the_hero != NULL);

	sspots = 0;
	for (y = 0; y < MAP_HEIGHT; y++)
	{
		for (x = 0; x < MAP_WIDTH; x++)
		{
			if ((!tiles[x][y].is_solid) &&
				(abs(x - the_game.the_hero->get_x()) > 10) && (abs(y - the_game.the_hero->get_y()) > 10))
			{
				sspots++;
				if (random(sspots) == 0)
				{
					sx = x;
					sy = y;
				}
			}
		}
	}

	if (sspots > 0)
	{
		// place the stairs
		tiles[sx][sy].display_ch = '>';
		tiles[sx][sy].is_stairs = true;
		tiles[sx][sy].display_style = mk_char_style(255, 255, 255, 64, 64, 64);

		if (level == LT_LAIR)
		{
			// place the witch above the stairs
			if (tiles[sx][sy].my_creature != NULL)
			{
				return false;
			}
			new creature(the_game, this, sx, sy, CT_WITCH, true);
		}

		return true;
	}
	
	return false;
}

bool map :: add_monster(game &the_game, level_type level)
{
	creature_type ct;
	int x, y, i;

	// choose a location
	x = random(1, MAP_WIDTH - 2);
	y = random(1, MAP_HEIGHT - 2);
	if ((tiles[x][y].is_solid) ||
		(tiles[x][y].my_creature != NULL))
	{
		return false;
	}

	// choose a typical creature type
	i = random(18);
	ct = CT_GOBLIN;											// common (4)
	if ((i == 4) || (i == 5))		{ct = CT_CENTIPEDE;}	// uncommon (2)
	if ((i == 6) || (i == 7))		{ct = CT_ANT;}			// uncommon (2)
	if ((i == 8) || (i == 9))		{ct = CT_SPIDER;}		// uncommon (2)
	if ((i == 10) || (i == 11))		{ct = CT_BEETLE;}		// uncommon (2)
	if (i == 12)					{ct = CT_MANTIS;}		// rare (1)
	if (level == LT_ENTRY)
	{
		// nothing particularly scary on this level
		if ((i >= 13) && (i <= 15))	{ct = CT_SNAKE;}		// common (3)
		if ((i == 16) || (i == 17))	{ct = CT_BEAR;}			// uncommon (2)
	} else {
		// some rougher monsters
		if (i == 13)					{ct = CT_HOBGOBLIN;}	// rare (1)
		if (i == 14)					{ct = CT_MANTICORE;}	// rare (1)
		if (i == 15)					{ct = CT_GRIFFON;}		// rare (1)
		if (i == 16)					{ct = CT_TROLL;}		// rare (1)
		if (i == 17)					{ct = CT_DRAGON;}		// rare (1)
	}

	switch (level)
	{
	case LT_ENTRY:
		{
		} break;

	case LT_TUNNELS:
		{
			if (random(8) == 0)
			{
				// lots of trolls
				ct = CT_TROLL;
			}
		} break;

	case LT_CRYPT:
		{
			if (random(8) == 0)
			{
				// lots of wraiths
				ct = CT_WRAITH;
			}
		} break;

	case LT_GOBLINHALLS:
		{
			if (random(2) == 0)
			{
				// lots of goblins and hobgoblins
				ct = (random(5) == 0) ? (CT_HOBGOBLIN) : (CT_GOBLIN);
			}
		} break;
		
	case LT_LAIR:
		{
		} break;

	default:
		{
			assert(false);
		} break;
	}

	// create the creature
	new creature(the_game, this, x, y, ct, true);
	return true;
}

int map :: make_connected(game &the_game)
{
	bool is_connected[MAP_WIDTH][MAP_HEIGHT];
	int x, y, area, prev_area;

	// start with only the hero connected
	for (y = 0; y < MAP_HEIGHT; y++)
	{
		for (x = 0; x < MAP_WIDTH; x++)
		{
			is_connected[x][y] = false;
		}
	}
	if ((the_game.the_hero != NULL) && (the_game.the_hero->get_map() == this))
	{
		is_connected[the_game.the_hero->get_x()][the_game.the_hero->get_y()] = true;
	}

	area = 0;
	prev_area = -1;
	while (area != prev_area)
	{
		// flood fill right and down
		for (y = 1; y < MAP_HEIGHT; y++)
		{
			for (x = 1; x < MAP_WIDTH; x++)
			{
				if ((!tiles[x][y].is_solid) &&
					(!is_connected[x][y]))
				{
					if ((is_connected[x - 1][y]) ||
						(is_connected[x][y - 1]) ||
						(is_connected[x - 1][y - 1]))
					{
						is_connected[x][y] = true;
					}
				}
			}
		}

		// flood fill left and down
		for (y = 1; y < MAP_HEIGHT; y++)
		{
			for (x = MAP_WIDTH - 2; x >= 0; x--)
			{
				if ((!tiles[x][y].is_solid) &&
					(!is_connected[x][y]))
				{
					if ((is_connected[x + 1][y]) ||
						(is_connected[x][y - 1]) ||
						(is_connected[x + 1][y - 1]))
					{
						is_connected[x][y] = true;
					}
				}
			}
		}

		// flood fill right and up
		for (y = MAP_HEIGHT - 2; y >= 0; y--)
		{
			for (x = 1; x < MAP_WIDTH; x++)
			{
				if ((!tiles[x][y].is_solid) &&
					(!is_connected[x][y]))
				{
					if ((is_connected[x - 1][y]) ||
						(is_connected[x][y + 1]) ||
						(is_connected[x - 1][y + 1]))
					{
						is_connected[x][y] = true;
					}
				}
			}
		}


		// flood fill left and up
		for (y = MAP_HEIGHT - 2; y >= 0; y--)
		{
			for (x = MAP_WIDTH - 2; x >= 0; x--)
			{
				if ((!tiles[x][y].is_solid) &&
					(!is_connected[x][y]))
				{
					if ((is_connected[x + 1][y]) ||
						(is_connected[x][y + 1]) ||
						(is_connected[x + 1][y + 1]))
					{
						is_connected[x][y] = true;
					}
				}
			}
		}

		// measure area
		prev_area = area;
		area = 0;
		for (y = 0; y < MAP_HEIGHT; y++)
		{
			for (x = 0; x < MAP_WIDTH; x++)
			{
				if (is_connected[x][y])
				{
					area++;
				}
			}
		}
	}

	// fill in any disconnected areas
	for (y = 0; y < MAP_HEIGHT; y++)
	{
		for (x = 0; x < MAP_WIDTH; x++)
		{
			if ((!tiles[x][y].is_solid) &&
				(!is_connected[x][y]))
			{
				tiles[x][y].display_ch = '#';
				tiles[x][y].is_solid = true;
			}
		}
	}

	return area;
}

bool map :: try_gen_level(game &the_game, level_type level, creature *the_hero)
{
	int x, y, i, area, monsters;

	assert(level >= 0);
	assert(level < LT_NUM);

	// remove the hero from his current location
	the_hero->move_to(NULL, 0, 0);

	// start with solid rock, nothing else
	clear_creatures();
	clear_items();
	for (y = 0; y < MAP_HEIGHT; y++)
	{
		for (x = 0; x < MAP_WIDTH; x++)
		{
			tiles[x][y].display_ch = '#';
			tiles[x][y].display_style = level_infos[level].rock_style;
			tiles[x][y].is_fx = false;
			tiles[x][y].is_solid = true;
			tiles[x][y].is_stairs = false;
			tiles[x][y].vis = V_HIDDEN;
			tiles[x][y].my_creature = NULL;
		}
	}

	// add rooms
	for (i = 0; i < level_infos[level].rooms_gen; i++)
	{
		add_room(level);
	}

	// add tunnels
	for (i = 0; i < level_infos[level].tunnels_gen; i++)
	{
		add_tunnel(level);
	}

	// add player
	if (!add_player(the_game, the_hero))
	{
		return false;
	}

	// connectedness processing
	area = make_connected(the_game);
#ifdef DEBUG_LEVELGEN
	char_debug_out("level area = " + char_int_str(area));
#endif // DEBUG_LEVELGEN
	if (area < level_infos[level].min_area)
	{
		return false;
	}

	// add stairs
	if (!add_stairs(the_game, level))
	{
		return false;
	}

	// add monsters
	monsters = random(level_infos[level].monsters_min, level_infos[level].monsters_max);
	i = 0;
	while (monsters > 0)
	{
		if (add_monster(the_game, level))
		{
			monsters--;
		} else if (++i > 1000) {
			// give up
			return false;
		}
	}

	return true;
}

void map :: gen_level(game &the_game, level_type _level, creature *the_hero)
{
	int attempts = 0;

	level = _level;

#ifdef DEBUG_LEVELGEN
	char_debug_out("generating level (" + char_int_str(_level) + ")");
#endif // DEBUG_LEVELGEN

	// keep trying to generate a level
	while (true)
	{
		attempts++;
		if (try_gen_level(the_game, level, the_hero))
		{
#ifdef DEBUG_LEVELGEN
			char_debug_out("success in " + char_int_str(attempts) + " attempts");
#endif // DEBUG_LEVELGEN
			return;
		}
	}
}

void map :: draw()
{
	int x, y;

	for (y = 0; y < MAP_HEIGHT; y++)
	{
		for (x = 0; x < MAP_WIDTH; x++)
		{
			if (tiles[x][y].vis == V_VISIBLE)
			{
				if ((tiles[x][y].my_items.empty()) ||
					(tiles[x][y].is_stairs))
				{
					// no items (or stairs, which takes priority)
					char_draw(x, y, tiles[x][y].display_ch, tiles[x][y].display_style);
				} else if (tiles[x][y].my_items.size() == 1) {
					// an item
					tiles[x][y].my_items.front()->draw(x, y);
				} else {
					// multiple items
					char_draw(x, y, '*', ITEM_DEFAULT_STYLE);
				}
#ifdef DEBUG_SEE_EVERYTHING
			} else if (true) {
#else
			} else if (tiles[x][y].vis == V_EXPLORED) {
#endif // DEBUG_SEE_EVERYTHING
				char_style fogged_style = tiles[x][y].display_style;

				// explored tiles are darker, slightly blue tinted
				// (the back colour is tinted but not darkened much, since it's usually dark anyway)
				fogged_style.fore_r = (fogged_style.fore_r * 3) / 8;
				fogged_style.fore_g = (fogged_style.fore_g * 3) / 8;
				fogged_style.fore_b = (fogged_style.fore_b * 4) / 8;
				fogged_style.back_r = (fogged_style.back_r * 6) / 8;
				fogged_style.back_g = (fogged_style.back_g * 6) / 8;
				fogged_style.back_b = (fogged_style.back_b * 8) / 8;

				char_draw(x, y, tiles[x][y].display_ch, fogged_style);
			} else {
				assert(tiles[x][y].vis == V_HIDDEN);
				char_draw(x, y, ' ', char_default_style);
			}
		}
	}
}

void map :: draw_fx()
{
	int x, y;

	for (y = 0; y < MAP_HEIGHT; y++)
	{
		for (x = 0; x < MAP_WIDTH; x++)
		{
#ifdef DEBUG_SEE_EVERYTHING
			if (tiles[x][y].is_fx)
#else
			if ((tiles[x][y].is_fx) && (tiles[x][y].vis == V_VISIBLE))
#endif // DEBUG_SEE_EVERYTHING
			{
				char_draw(x, y, tiles[x][y].fx_ch, tiles[x][y].fx_style);
			}
		}
	}
}

map_tile *map :: get_tile(int x, int y)
{
	assert((x >= 0) && (y >= 0) && (x < MAP_WIDTH) && (y < MAP_HEIGHT));

	return &(tiles[x][y]);
}

map_tile *map :: get_tile(map *_map, int x, int y)
{
	if ((_map != NULL) &&
		(x >= 0) && (y >= 0) && (x < MAP_WIDTH) && (y < MAP_HEIGHT))
	{
		return _map->get_tile(x, y);
	} else {
		return NULL;
	}
}

std::string map :: get_level_introduction()
{
	assert(level >= 0);
	assert(level < LT_NUM);

	return level_infos[level].introduction;
}